package com.ttu.naabrivalve.agent;

import com.ttu.naabrivalve.model.GPSPoint;
import com.ttu.naabrivalve.model.Home;
import com.ttu.naabrivalve.model.Location;
import com.ttu.naabrivalve.onto.*;
import com.ttu.naabrivalve.service.GPSService;
import com.ttu.naabrivalve.service.LocationService;
import jade.content.Concept;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.content.onto.basic.Result;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.core.behaviours.TickerBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.domain.FIPAException;
import jade.lang.acl.ACLMessage;

import java.util.Date;
import java.util.Random;

/**
 * Created by ragnarp on 25.04.2015.
 */
public class CrimeWatchAgent extends Agent implements CrimeWatchVocabulary {

    private Codec codec = new SLCodec();
    private Ontology ontology = CrimeWatchOntology.getInstance();
    private Logger logger = new Logger(1);
    private LocationService gps;
    private Home home;
    private boolean atHome;

    {
        gps = new GPSService();
    }

    protected void setup() {
        getContentManager().registerLanguage(codec);
        getContentManager().registerOntology(ontology);

        try {
            Thread.sleep(new Random().nextInt(10000) + 1000);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        logger.info(String.format("[%s] Tere! CrimeWatch agent alustab tööd.", getAID().getName()));

        Object args[] = getArguments();
        if (args != null && args.length > 0) {}

        this.home = new Home("Kodu1", new GPSPoint(0.0, 0.0));
        this.atHome = true;

        DFAgentDescription dfAgentDescription = new DFAgentDescription();
        dfAgentDescription.setName(getAID());

        ServiceDescription outOfHome = new ServiceDescription();
        outOfHome.setType("out-of-home");
        outOfHome.setName("JADE-crime-watch");

        dfAgentDescription.addServices(outOfHome);

        try {
            DFService.register(this, dfAgentDescription);

        } catch (FIPAException e) {
            e.printStackTrace();
        }

        logger.info(String.format("[%s] Sõnumite vastuvõtmise sisselülitamine...", getAID().getName()));
        addBehaviour(new CyclicBehaviour() {
            @Override
            public void action() {
                ACLMessage aclMessage = myAgent.receive();

                if (aclMessage != null) {
                    try {
                        ContentElement contentElement = getContentManager().extractContent(aclMessage);

                        switch (aclMessage.getPerformative()) {
                            case (ACLMessage.REQUEST):
                                Concept action = ((Action) contentElement).getAction();
                                logger.info("[" + getAID().getLocalName() + "]: "
                                        + action.getClass().getName()
                                        + ": " + aclMessage.getSender().getLocalName());
                                ((CrimeWatchAction) action).handle(aclMessage, myAgent);
                                break;
                            case (ACLMessage.INFORM):
                                if (contentElement instanceof Result) {
                                    Result result = (Result) contentElement;
                                    if (result.getValue() instanceof Message) {
                                        Message msg = (Message) result.getValue();
                                        logger.info(
                                                "[" + getAID().getLocalName() + "]: "
                                                        + "INFORM from " + aclMessage.getSender().getLocalName()
                                                        + ", message: " + msg.getContent()
                                        );
                                    }
                                }
                                break;
                            default:
                                replyNotUnderstood(aclMessage);
                        }
                    } catch (Codec.CodecException e) {
                        e.printStackTrace();
                    } catch (OntologyException e) {
                        e.printStackTrace();
                    }
                } else {
                    block();
                }
            }
        });

        logger.info(String.format("[%s] Asukoha monitooringu sisselülitamine...%n", getAID().getName()));
        addBehaviour(new TickerBehaviour(this, 10000) {

            private AID[] findPeerAgents(){
                // Leia kõik agendid, kes pakuvad "kodust väljas" teavitamise teenust
                DFAgentDescription template = new DFAgentDescription();
                ServiceDescription serviceDescription1 = new ServiceDescription();
                serviceDescription1.setType("out-of-home");
                template.addServices(serviceDescription1);

                try {
                    DFAgentDescription[] dfAgentDescriptions = DFService.search(myAgent, template);
                    AID[] peerAgents = new AID[dfAgentDescriptions.length];
                    for (int i = 0; i < dfAgentDescriptions.length; i++) {
                        peerAgents[i] = dfAgentDescriptions[i].getName();
                    }
                    return peerAgents;
                } catch (FIPAException e) {
                    e.printStackTrace();
                }

                return new AID[]{};
            }

            private void notifyPeers(CrimeWatchAction action){
                // Teavita partneragente
                AID[] peerAgents = findPeerAgents();
                for (int i = 0; i < peerAgents.length; i++) {
                    // Jäta iseennast vahele
                    if (peerAgents[i].equals(getAID())) continue;

                    ACLMessage aclMessage = new ACLMessage(ACLMessage.REQUEST);
                    aclMessage.setLanguage(codec.getName());
                    aclMessage.setOntology(ontology.getName());
                    try {
                        getContentManager().fillContent(aclMessage, new Action(peerAgents[i], action));
                    } catch (Codec.CodecException e) {
                        e.printStackTrace();
                    } catch (OntologyException e) {
                        e.printStackTrace();
                    }
                    aclMessage.addReceiver(peerAgents[i]);
                    myAgent.send(aclMessage);
                }
            }

            @Override
            protected void onTick() {
                Location currentLocation = gps.getCurrentLocation();

                if (!(currentLocation.getCoordinates().equals(home.getCoordinates()))) {
                    // Kodust väljas ...
                    RegisterOutOfHome outOfHome = new RegisterOutOfHome();
                    outOfHome.setWhen(new Date());
                    notifyPeers(outOfHome);
                    atHome = false;
                } else {
                    // Kui just jõuti koju ...
                    if (!atHome){
                        RegisterBackAtHome backAtHome = new RegisterBackAtHome();
                        backAtHome.setWhen(new Date());
                        notifyPeers(backAtHome);
                    }
                    atHome = true;
                }
            }
        });
    }

    protected void replyNotUnderstood(ACLMessage aclMessage){
        try {
            ContentElement contentElement = getContentManager().extractContent(aclMessage);
            ACLMessage reply = aclMessage.createReply();
            reply.setPerformative(ACLMessage.NOT_UNDERSTOOD);
            getContentManager().fillContent(reply, contentElement);
            send(reply);
        } catch (Codec.CodecException e) {
            e.printStackTrace();
        } catch (OntologyException e) {
            e.printStackTrace();
        }

    }

    @Override
    protected void takeDown() {
        try {
            // Lõpeta teenuste pakkumine ...
            DFService.deregister(this);
        } catch (FIPAException e) {
            e.printStackTrace();
        }

        logger.info(String.format("[%s] agent %s lõpetab.", getAID().getName(), getAID().getName()));
    }

    private class Logger {

        int level;

        Logger(int level) {
            this.level = level;
        }

        void info(Object o){
            if (level >= 1)
                log(o, "INFO");
        }
        void debug(Object o){
            if (level >= 2)
                log(o, "DEBUG");
        }

        private void log(Object o, String LEVEL){
            System.out.println(LEVEL + ";" + new Date().toString() + ";" + o.toString());
        }
    }
}
